@@ -34,95 +34,149 @@ PKCS11_CTX *PKCS11_CTX_new(void)
 
 void PKCS11_CTX_init_args(PKCS11_CTX *ctx, const char *init_args)
 {
-	if (check_fork(ctx) < 0)
+	if (check_fork(PRIVCTX(ctx)) < 0)
 		return;
 	pkcs11_CTX_init_args(ctx, init_args);
 }
 
 int PKCS11_CTX_load(PKCS11_CTX *ctx, const char *ident)
 {
-	if (check_fork(ctx) < 0)
+	if (check_fork(PRIVCTX(ctx)) < 0)
 		return -1;
 	return pkcs11_CTX_load(ctx, ident);
 }
 
 void PKCS11_CTX_unload(PKCS11_CTX *ctx)
 {
-	if (check_fork(ctx) < 0)
+	if (check_fork(PRIVCTX(ctx)) < 0)
 		return;
 	pkcs11_CTX_unload(ctx);
 }
 
 void PKCS11_CTX_free(PKCS11_CTX *ctx)
 {
-	if (check_fork(ctx) < 0)
+	if (check_fork(PRIVCTX(ctx)) < 0)
 		return;
 	pkcs11_CTX_free(ctx);
 }
 
-int PKCS11_open_session(PKCS11_SLOT *slot, int rw)
+int PKCS11_open_session(PKCS11_SLOT *pslot, int rw)
 {
+	PKCS11_SLOT_private *slot = PRIVSLOT(pslot);
 	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_open_session(slot, rw, 0);
+	return pkcs11_open_session(slot, rw);
 }
 
-int PKCS11_enumerate_slots(PKCS11_CTX *ctx,
+int PKCS11_enumerate_slots(PKCS11_CTX *pctx,
 		PKCS11_SLOT **slotsp, unsigned int *nslotsp)
 {
+	PKCS11_CTX_private *ctx = PRIVCTX(pctx);
 	if (check_fork(ctx) < 0)
 		return -1;
+	if (!nslotsp)
+		return -1;
+	if (slotsp)
+		*slotsp = 0;
+	if (nslotsp)
+		*nslotsp = 0;
+	return pkcs11_enumerate_slots(ctx, slotsp, nslotsp);
+}
+
+int PKCS11_update_slots(PKCS11_CTX *pctx,
+		PKCS11_SLOT **slotsp, unsigned int *nslotsp)
+{
+	PKCS11_CTX_private *ctx = PRIVCTX(pctx);
+	if (check_fork(ctx) < 0)
+		return -1;
+	if (!nslotsp)
+		return -1;
 	return pkcs11_enumerate_slots(ctx, slotsp, nslotsp);
 }
 
-unsigned long PKCS11_get_slotid_from_slot(PKCS11_SLOT *slot)
+unsigned long PKCS11_get_slotid_from_slot(PKCS11_SLOT *pslot)
 {
+	PKCS11_SLOT_private *slot = PRIVSLOT(pslot);
 	if (check_slot_fork(slot) < 0)
 		return 0L;
 	return pkcs11_get_slotid_from_slot(slot);
 }
 
-void PKCS11_release_all_slots(PKCS11_CTX *ctx,
+void PKCS11_release_all_slots(PKCS11_CTX *pctx,
 		PKCS11_SLOT *slots, unsigned int nslots)
 {
+	PKCS11_CTX_private *ctx = PRIVCTX(pctx);
 	if (check_fork(ctx) < 0)
 		return;
-	pkcs11_release_all_slots(ctx, slots, nslots);
+	pkcs11_release_all_slots(slots, nslots);
 }
 
 PKCS11_SLOT *PKCS11_find_token(PKCS11_CTX *ctx,
 		PKCS11_SLOT *slots, unsigned int nslots)
 {
-	if (check_fork(ctx) < 0)
+	PKCS11_SLOT *slot, *best;
+	PKCS11_TOKEN *tok;
+	unsigned int n;
+
+	if (check_fork(PRIVCTX(ctx)) < 0)
 		return NULL;
-	return pkcs11_find_token(ctx, slots, nslots);
+	if (!slots)
+		return NULL;
+
+	best = NULL;
+	for (n = 0, slot = slots; n < nslots; n++, slot++) {
+		if ((tok = slot->token) != NULL) {
+			if (!best ||
+					(tok->initialized > best->token->initialized &&
+					tok->userPinSet > best->token->userPinSet &&
+					tok->loginRequired > best->token->loginRequired))
+				best = slot;
+		}
+	}
+	return best;
 }
 
 PKCS11_SLOT *PKCS11_find_next_token(PKCS11_CTX *ctx,
 		PKCS11_SLOT *slots, unsigned int nslots,
 		PKCS11_SLOT *current)
 {
-	if (check_fork(ctx) < 0)
+	int offset;
+
+	if (check_fork(PRIVCTX(ctx)) < 0)
 		return NULL;
-	return pkcs11_find_next_token(ctx, slots, nslots, current);
+	if (!slots)
+		return NULL;
+
+	if (current) {
+		offset = current + 1 - slots;
+		if (offset < 1 || (unsigned int)offset >= nslots)
+			return NULL;
+	} else {
+		offset = 0;
+	}
+
+	return PKCS11_find_token(ctx, slots + offset, nslots - offset);
 }
 
-int PKCS11_is_logged_in(PKCS11_SLOT *slot, int so, int *res)
+int PKCS11_is_logged_in(PKCS11_SLOT *pslot, int so, int *res)
 {
+	PKCS11_SLOT_private *slot = PRIVSLOT(pslot);
 	if (check_slot_fork(slot) < 0)
 		return -1;
 	return pkcs11_is_logged_in(slot, so, res);
 }
 
-int PKCS11_login(PKCS11_SLOT *slot, int so, const char *pin)
+int PKCS11_login(PKCS11_SLOT *pslot, int so, const char *pin)
 {
+	PKCS11_SLOT_private *slot = PRIVSLOT(pslot);
 	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_login(slot, so, pin, 0);
+	return pkcs11_login(slot, so, pin);
 }
 
-int PKCS11_logout(PKCS11_SLOT *slot)
+int PKCS11_logout(PKCS11_SLOT *pslot)
 {
+	PKCS11_SLOT_private *slot = PRIVSLOT(pslot);
 	if (check_slot_fork(slot) < 0)
 		return -1;
 	return pkcs11_logout(slot);
@@ -131,57 +185,65 @@ int PKCS11_logout(PKCS11_SLOT *slot)
 int PKCS11_enumerate_keys(PKCS11_TOKEN *token,
 		PKCS11_KEY **keys, unsigned int *nkeys)
 {
-	if (check_token_fork(token) < 0)
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
+	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_enumerate_keys(token, CKO_PRIVATE_KEY, keys, nkeys);
+	return pkcs11_enumerate_keys(slot, CKO_PRIVATE_KEY, keys, nkeys);
 }
 
-int PKCS11_remove_key(PKCS11_KEY *key)
+int PKCS11_remove_key(PKCS11_KEY *pkey)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return -1;
-	return pkcs11_remove_key(key);
+	return pkcs11_remove_object(key);
 }
 
 int PKCS11_enumerate_public_keys(PKCS11_TOKEN *token,
 		PKCS11_KEY **keys, unsigned int *nkeys)
 {
-	if (check_token_fork(token) < 0)
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
+	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_enumerate_keys(token, CKO_PUBLIC_KEY, keys, nkeys);
+	return pkcs11_enumerate_keys(slot, CKO_PUBLIC_KEY, keys, nkeys);
 }
 
-int PKCS11_get_key_type(PKCS11_KEY *key)
+int PKCS11_get_key_type(PKCS11_KEY *pkey)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return -1;
 	return pkcs11_get_key_type(key);
 }
 
-EVP_PKEY *PKCS11_get_private_key(PKCS11_KEY *key)
+EVP_PKEY *PKCS11_get_private_key(PKCS11_KEY *pkey)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return NULL;
-	return pkcs11_get_key(key, 1);
+	return pkcs11_get_key(key, CKO_PRIVATE_KEY);
 }
 
-EVP_PKEY *PKCS11_get_public_key(PKCS11_KEY *key)
+EVP_PKEY *PKCS11_get_public_key(PKCS11_KEY *pkey)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return NULL;
-	return pkcs11_get_key(key, 0);
+	return pkcs11_get_key(key, CKO_PUBLIC_KEY);
 }
 
-PKCS11_CERT *PKCS11_find_certificate(PKCS11_KEY *key)
+PKCS11_CERT *PKCS11_find_certificate(PKCS11_KEY *pkey)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return NULL;
 	return pkcs11_find_certificate(key);
 }
 
-PKCS11_KEY *PKCS11_find_key(PKCS11_CERT *cert)
+PKCS11_KEY *PKCS11_find_key(PKCS11_CERT *pcert)
 {
-	if (check_cert_fork(cert) < 0)
+	PKCS11_OBJECT_private *cert = PRIVCERT(pcert);
+	if (check_object_fork(cert) < 0)
 		return NULL;
 	return pkcs11_find_key(cert);
 }
@@ -189,75 +251,100 @@ PKCS11_KEY *PKCS11_find_key(PKCS11_CERT *cert)
 int PKCS11_enumerate_certs(PKCS11_TOKEN *token,
 		PKCS11_CERT **certs, unsigned int *ncerts)
 {
-	if (check_token_fork(token) < 0)
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
+	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_enumerate_certs(token, certs, ncerts);
+	return pkcs11_enumerate_certs(slot, certs, ncerts);
 }
 
-int PKCS11_remove_certificate(PKCS11_CERT *cert)
+int PKCS11_remove_certificate(PKCS11_CERT *pcert)
 {
-	if (check_cert_fork(cert) < 0)
+	PKCS11_OBJECT_private *cert = PRIVCERT(pcert);
+	if (check_object_fork(cert) < 0)
 		return -1;
-	return pkcs11_remove_certificate(cert);
+	return pkcs11_remove_object(cert);
 }
 
 int PKCS11_init_token(PKCS11_TOKEN *token, const char *pin,
 		const char *label)
 {
-	if (check_token_fork(token) < 0)
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
+	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_init_token(token, pin, label);
+	return pkcs11_init_token(slot, pin, label);
 }
 
 int PKCS11_init_pin(PKCS11_TOKEN *token, const char *pin)
 {
-	if (check_token_fork(token) < 0)
-		return -1;
-	return pkcs11_init_pin(token, pin);
-}
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
+	int r;
 
-int PKCS11_change_pin(PKCS11_SLOT *slot,
-		const char *old_pin, const char *new_pin)
-{
 	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_change_pin(slot, old_pin, new_pin);
+	r = pkcs11_init_pin(slot, pin);
+	if (r == 0)
+		r = pkcs11_refresh_token(token->slot);
+	return r;
+}
+
+int PKCS11_change_pin(PKCS11_SLOT *pslot,
+		const char *old_pin, const char *new_pin)
+{
+	PKCS11_SLOT_private *slot = PRIVSLOT(pslot);
+	int r;
+
+	if (check_slot_fork(slot) < 0)
+		return -1;
+	r = pkcs11_change_pin(slot, old_pin, new_pin);
+	if (r == 0)
+		r = pkcs11_refresh_token(pslot);
+	return r;
 }
 
 int PKCS11_store_private_key(PKCS11_TOKEN *token,
 		EVP_PKEY *pk, char *label, unsigned char *id, size_t id_len)
 {
-	if (check_token_fork(token) < 0)
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
+	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_store_private_key(token, pk, label, id, id_len);
+	return pkcs11_store_private_key(slot, pk, label, id, id_len);
 }
 
 int PKCS11_store_public_key(PKCS11_TOKEN *token,
     	EVP_PKEY *pk, char *label, unsigned char *id, size_t id_len)
 {
-	if (check_token_fork(token) < 0)
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
+	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_store_public_key(token, pk, label, id, id_len);
+	return pkcs11_store_public_key(slot, pk, label, id, id_len);
 }
 
 int PKCS11_store_certificate(PKCS11_TOKEN *token, X509 *x509,
 		char *label, unsigned char *id, size_t id_len,
 		PKCS11_CERT **ret_cert)
 {
-	if (check_token_fork(token) < 0)
-		return -1;
-	return pkcs11_store_certificate(token, x509, label, id, id_len, ret_cert);
-}
-
-int PKCS11_seed_random(PKCS11_SLOT *slot, const unsigned char *s, unsigned int s_len)
-{
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
 	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_seed_random(slot, s, s_len);
+	return pkcs11_store_certificate(slot, x509, label, id, id_len, ret_cert);
 }
 
-int PKCS11_generate_random(PKCS11_SLOT *slot, unsigned char *r, unsigned int r_len)
+int PKCS11_seed_random(PKCS11_SLOT *pslot, const unsigned char *s, unsigned int s_len)
 {
+	PKCS11_SLOT_private *slot = PRIVSLOT(pslot);
+	int r;
+
+	if (check_slot_fork(slot) < 0)
+		return -1;
+	r = pkcs11_seed_random(slot, s, s_len);
+	if (r == 0)
+		r = pkcs11_refresh_token(pslot);
+	return r;
+}
+
+int PKCS11_generate_random(PKCS11_SLOT *pslot, unsigned char *r, unsigned int r_len)
+{
+	PKCS11_SLOT_private *slot = PRIVSLOT(pslot);
 	if (check_slot_fork(slot) < 0)
 		return -1;
 	return pkcs11_generate_random(slot, r, r_len);
@@ -272,8 +359,9 @@ void ERR_load_PKCS11_strings(void)
 	ERR_load_CKR_strings();
 }
 
-int PKCS11_set_ui_method(PKCS11_CTX *ctx, UI_METHOD *ui_method, void *ui_user_data)
+int PKCS11_set_ui_method(PKCS11_CTX *pctx, UI_METHOD *ui_method, void *ui_user_data)
 {
+	PKCS11_CTX_private *ctx = PRIVCTX(pctx);
 	if (check_fork(ctx) < 0)
 		return -1;
 	return pkcs11_set_ui_method(ctx, ui_method, ui_user_data);
@@ -285,52 +373,59 @@ int PKCS11_generate_key(PKCS11_TOKEN *token,
 		int algorithm, unsigned int bits,
 		char *label, unsigned char *id, size_t id_len)
 {
-	if (check_token_fork(token) < 0)
+	PKCS11_SLOT_private *slot = PRIVSLOT(token->slot);
+	if (check_slot_fork(slot) < 0)
 		return -1;
-	return pkcs11_generate_key(token, algorithm, bits, label, id, id_len);
+	return pkcs11_generate_key(slot, algorithm, bits, label, id, id_len);
 }
 
-int PKCS11_get_key_size(PKCS11_KEY *key)
+int PKCS11_get_key_size(PKCS11_KEY *pkey)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return -1;
 	return pkcs11_get_key_size(key);
 }
 
-int PKCS11_get_key_modulus(PKCS11_KEY *key, BIGNUM **bn)
+int PKCS11_get_key_modulus(PKCS11_KEY *pkey, BIGNUM **bn)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return -1;
 	return pkcs11_get_key_modulus(key, bn);
 }
 
-int PKCS11_get_key_exponent(PKCS11_KEY *key, BIGNUM **bn)
+int PKCS11_get_key_exponent(PKCS11_KEY *pkey, BIGNUM **bn)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return -1;
 	return pkcs11_get_key_exponent(key, bn);
 }
 
 int PKCS11_sign(int type, const unsigned char *m, unsigned int m_len,
-		unsigned char *sigret, unsigned int *siglen, PKCS11_KEY *key)
+		unsigned char *sigret, unsigned int *siglen, PKCS11_KEY *pkey)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return -1;
 	return pkcs11_sign(type, m, m_len, sigret, siglen, key);
 }
 
 int PKCS11_private_encrypt(int flen, const unsigned char *from, unsigned char *to,
-		PKCS11_KEY *key, int padding)
+		PKCS11_KEY *pkey, int padding)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return -1;
 	return pkcs11_private_encrypt(flen, from, to, key, padding);
 }
 
 int PKCS11_private_decrypt(int flen, const unsigned char *from, unsigned char *to,
-		PKCS11_KEY *key, int padding)
+		PKCS11_KEY *pkey, int padding)
 {
-	if (check_key_fork(key) < 0)
+	PKCS11_OBJECT_private *key = PRIVKEY(pkey);
+	if (check_object_fork(key) < 0)
 		return -1;
 	return pkcs11_private_decrypt(flen, from, to, key, padding);
 }
