@@ -40,9 +40,7 @@ PKCS11_CTX *pkcs11_CTX_new(void)
 	memset(ctx, 0, sizeof(PKCS11_CTX));
 	ctx->_private = cpriv;
 	cpriv->forkid = get_forkid();
-	cpriv->rwlock = CRYPTO_THREAD_lock_new();
-	cpriv->sign_initialized = 0;
-	cpriv->decrypt_initialized = 0;
+	pthread_mutex_init(&cpriv->fork_lock, 0);
 
 	return ctx;
 fail:
@@ -113,23 +111,22 @@ int pkcs11_CTX_load(PKCS11_CTX *ctx, const char *name)
 /*
  * Reinitialize (e.g., after a fork).
  */
-int pkcs11_CTX_reload(PKCS11_CTX *ctx)
+int pkcs11_CTX_reload(PKCS11_CTX_private *ctx)
 {
-	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
 	CK_C_INITIALIZE_ARGS _args;
 	CK_C_INITIALIZE_ARGS *args = NULL;
 	int rv;
 
-	if (!cpriv->method) /* Module not loaded */
+	if (!ctx->method) /* Module not loaded */
 		return 0;
 
 	/* Tell the PKCS11 to initialize itself */
-	if (cpriv->init_args) {
+	if (ctx->init_args) {
 		memset(&_args, 0, sizeof(_args));
 		args = &_args;
-		args->pReserved = cpriv->init_args;
+		args->pReserved = ctx->init_args;
 	}
-	rv = cpriv->method->C_Initialize(args);
+	rv = ctx->method->C_Initialize(args);
 	if (rv && rv != CKR_CRYPTOKI_ALREADY_INITIALIZED) {
 		CKRerr(P11_F_PKCS11_CTX_RELOAD, rv);
 		return -1;
@@ -172,7 +169,7 @@ void pkcs11_CTX_free(PKCS11_CTX *ctx)
 	if (cpriv->handle) {
 		OPENSSL_free(cpriv->handle);
 	}
-	CRYPTO_THREAD_lock_free(cpriv->rwlock);
+	pthread_mutex_destroy(&cpriv->fork_lock);
 	OPENSSL_free(ctx->manufacturer);
 	OPENSSL_free(ctx->description);
 	OPENSSL_free(ctx->_private);
