@@ -36,7 +36,6 @@ static int (*orig_pkey_ec_sign) (EVP_PKEY_CTX *ctx,
 	const unsigned char *tbs, size_t tbslen);
 #endif /* OPENSSL_NO_EC */
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
 struct evp_pkey_method_st {
 	int pkey_id;
 	int flags;
@@ -75,6 +74,9 @@ struct evp_pkey_method_st {
 	int (*ctrl) (EVP_PKEY_CTX *ctx, int type, int p1, void *p2);
 	int (*ctrl_str) (EVP_PKEY_CTX *ctx, const char *type, const char *value);
 } /* EVP_PKEY_METHOD */ ;
+
+#if OPENSSL_VERSION_NUMBER >= 0x10000000L
+#define EVP_PKEY_FLAG_DYNAMIC 1
 #endif
 
 #if OPENSSL_VERSION_NUMBER < 0x10002000L || defined(LIBRESSL_VERSION_NUMBER)
@@ -90,19 +92,27 @@ typedef struct {
 	unsigned char *tbuf;
 } RSA_PKEY_CTX;
 
+#endif
+
+#if OPENSSL_VERSION_NUMBER < 0x10002000L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x3000100L )
+
 static int EVP_PKEY_CTX_get_signature_md(EVP_PKEY_CTX *ctx, const EVP_MD **pmd)
 {
 	RSA_PKEY_CTX *rctx = EVP_PKEY_CTX_get_data(ctx);
-	if (rctx == NULL)
+	if (!rctx)
 		return -1;
 	*pmd = rctx->md;
 	return 1;
 }
 
+#endif
+
+#if OPENSSL_VERSION_NUMBER < 0x10002000L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x3010000L )
+
 static int EVP_PKEY_CTX_get_rsa_oaep_md(EVP_PKEY_CTX *ctx, const EVP_MD **pmd)
 {
 	RSA_PKEY_CTX *rctx = EVP_PKEY_CTX_get_data(ctx);
-	if (rctx == NULL)
+	if (!rctx)
 		return -1;
 	*pmd = rctx->md;
 	return 1;
@@ -115,7 +125,7 @@ static int EVP_PKEY_CTX_get_rsa_oaep_md(EVP_PKEY_CTX *ctx, const EVP_MD **pmd)
 static int EVP_PKEY_CTX_get_rsa_mgf1_md(EVP_PKEY_CTX *ctx, const EVP_MD **pmd)
 {
 	RSA_PKEY_CTX *rctx = EVP_PKEY_CTX_get_data(ctx);
-	if (rctx == NULL)
+	if (!rctx)
 		return -1;
 	*pmd = rctx->mgf1md;
 	return 1;
@@ -124,7 +134,7 @@ static int EVP_PKEY_CTX_get_rsa_mgf1_md(EVP_PKEY_CTX *ctx, const EVP_MD **pmd)
 static int EVP_PKEY_CTX_get_rsa_padding(EVP_PKEY_CTX *ctx, int *padding)
 {
 	RSA_PKEY_CTX *rctx = EVP_PKEY_CTX_get_data(ctx);
-	if (rctx == NULL)
+	if (!rctx)
 		return -1;
 	*padding = rctx->pad_mode;
 	return 1;
@@ -133,7 +143,7 @@ static int EVP_PKEY_CTX_get_rsa_padding(EVP_PKEY_CTX *ctx, int *padding)
 static int EVP_PKEY_CTX_get_rsa_pss_saltlen(EVP_PKEY_CTX *ctx, int *saltlen)
 {
 	RSA_PKEY_CTX *rctx = EVP_PKEY_CTX_get_data(ctx);
-	if (rctx == NULL)
+	if (!rctx)
 		return -1;
 	*saltlen = rctx->saltlen;
 	return 1;
@@ -187,6 +197,16 @@ static CK_MECHANISM_TYPE pkcs11_md2ckm(const EVP_MD *md)
 		return CKM_SHA512;
 	case NID_sha384:
 		return CKM_SHA384;
+#if !defined(LIBRESSL_VERSION_NUMBER)
+	case NID_sha3_224:
+		return CKM_SHA3_224;
+	case NID_sha3_256:
+		return CKM_SHA3_256;
+	case NID_sha3_384:
+		return CKM_SHA3_384;
+	case NID_sha3_512:
+		return CKM_SHA3_512;
+#endif
 	default:
 		return 0;
 	}
@@ -205,6 +225,16 @@ static CK_RSA_PKCS_MGF_TYPE pkcs11_md2ckg(const EVP_MD *md)
 		return CKG_MGF1_SHA512;
 	case NID_sha384:
 		return CKG_MGF1_SHA384;
+#if !defined(LIBRESSL_VERSION_NUMBER)
+	case NID_sha3_224:
+		return CKG_MGF1_SHA3_224;
+	case NID_sha3_256:
+		return CKG_MGF1_SHA3_256;
+	case NID_sha3_384:
+		return CKG_MGF1_SHA3_384;
+	case NID_sha3_512:
+		return CKG_MGF1_SHA3_512;
+#endif
 	default:
 		return 0;
 	}
@@ -230,7 +260,7 @@ static int pkcs11_params_pss(CK_RSA_PKCS_PSS_PARAMS *pss,
 		break;
 	case -2:
 		evp_pkey = EVP_PKEY_CTX_get0_pkey(ctx);
-		if (evp_pkey == NULL)
+		if (!evp_pkey)
 			return -1;
 		salt_len = EVP_PKEY_size(evp_pkey) - EVP_MD_size(sig_md) - 2;
 		if (((EVP_PKEY_bits(evp_pkey) - 1) & 0x7) == 0)
@@ -275,8 +305,8 @@ static int pkcs11_params_oaep(CK_RSA_PKCS_OAEP_PARAMS *oaep,
 	if (!oaep->hashAlg || !oaep->mgf)
 		return -1;
 	/* we do not support the OAEP "label" parameter yet... */
-	oaep->source = 0UL; /* empty parameter (label) */
-	oaep->pSourceData = NULL;
+	oaep->source = CKZ_DATA_SPECIFIED;
+	oaep->pSourceData = NULL; /* empty parameter (label) */
 	oaep->ulSourceDataLen = 0;
 	return 0;
 }
@@ -287,82 +317,79 @@ static int pkcs11_try_pkey_rsa_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 {
 	EVP_PKEY *pkey;
 	RSA *rsa;
-	PKCS11_KEY *key;
-	int rv = 0;
+	int rv = 0, padding;
 	CK_ULONG size = *siglen;
-	PKCS11_SLOT *slot;
-	PKCS11_CTX *ctx;
-	PKCS11_KEY_private *kpriv;
-	PKCS11_SLOT_private *spriv;
-	PKCS11_CTX_private *cpriv;
+	PKCS11_OBJECT_private *key;
+	PKCS11_SLOT_private *slot;
+	PKCS11_CTX_private *ctx;
 	const EVP_MD *sig_md;
+	CK_SESSION_HANDLE session;
+	CK_MECHANISM mechanism;
+	CK_RSA_PKCS_PSS_PARAMS pss_params;
 
 #ifdef DEBUG
 	fprintf(stderr, "%s:%d pkcs11_try_pkey_rsa_sign() "
 		"sig=%p *siglen=%lu tbs=%p tbslen=%lu\n",
 		__FILE__, __LINE__, sig, *siglen, tbs, tbslen);
 #endif
-	pkey = EVP_PKEY_CTX_get0_pkey(evp_pkey_ctx);
-	if (pkey == NULL)
+	/* RSA method has EVP_PKEY_FLAG_AUTOARGLEN set. OpenSSL core will handle
+	 * the size inquiry internally. */
+	 if (!sig)
 		return -1;
-	rsa = EVP_PKEY_get0_RSA(pkey);
-	if (rsa == NULL)
-		return -1;
-	key = pkcs11_get_ex_data_rsa(rsa);
-	if (check_key_fork(key) < 0)
-		return -1;
-	slot = KEY2SLOT(key);
-	ctx = KEY2CTX(key);
-	kpriv = PRIVKEY(key);
-	spriv = PRIVSLOT(slot);
-	cpriv = PRIVCTX(ctx);
 
-	if (evp_pkey_ctx == NULL)
+	pkey = EVP_PKEY_CTX_get0_pkey(evp_pkey_ctx);
+	if (!pkey)
+		return -1;
+	rsa = (RSA *)EVP_PKEY_get0_RSA(pkey);
+	if (!rsa)
+		return -1;
+
+	key = pkcs11_get_ex_data_rsa(rsa);
+	if (check_object_fork(key) < 0)
+		return -1;
+
+	slot = key->slot;
+	ctx = slot->ctx;
+	if (!evp_pkey_ctx)
 		return -1;
 	if (EVP_PKEY_CTX_get_signature_md(evp_pkey_ctx, &sig_md) <= 0)
 		return -1;
 	if (tbslen != (size_t)EVP_MD_size(sig_md))
 		return -1;
 
-	if (!cpriv->sign_initialized) {
-		int padding;
-		CK_MECHANISM mechanism;
-		CK_RSA_PKCS_PSS_PARAMS pss_params;
-
-		memset(&mechanism, 0, sizeof mechanism);
-		EVP_PKEY_CTX_get_rsa_padding(evp_pkey_ctx, &padding);
-		switch (padding) {
-		case RSA_PKCS1_PSS_PADDING:
+	memset(&mechanism, 0, sizeof mechanism);
+	EVP_PKEY_CTX_get_rsa_padding(evp_pkey_ctx, &padding);
+	switch (padding) {
+	case RSA_PKCS1_PSS_PADDING:
 #ifdef DEBUG
-			fprintf(stderr, "%s:%d padding=RSA_PKCS1_PSS_PADDING\n",
-				__FILE__, __LINE__);
-#endif
-			if (pkcs11_params_pss(&pss_params, evp_pkey_ctx) < 0)
-				return -1;
-			mechanism.mechanism = CKM_RSA_PKCS_PSS;
-			mechanism.pParameter = &pss_params;
-			mechanism.ulParameterLen = sizeof pss_params;
-			break;
-		default:
-#ifdef DEBUG
-			fprintf(stderr, "%s:%d unsupported padding: %d\n",
-				__FILE__, __LINE__, padding);
+		fprintf(stderr, "%s:%d padding=RSA_PKCS1_PSS_PADDING\n",
+			__FILE__, __LINE__);
 #endif
+		if (pkcs11_params_pss(&pss_params, evp_pkey_ctx) < 0)
 			return -1;
-		} /* end switch(padding) */
+		mechanism.mechanism = CKM_RSA_PKCS_PSS;
+		mechanism.pParameter = &pss_params;
+		mechanism.ulParameterLen = sizeof pss_params;
+		break;
+	default:
+#ifdef DEBUG
+		fprintf(stderr, "%s:%d unsupported padding: %d\n",
+			__FILE__, __LINE__, padding);
+#endif
+		return -1;
+	} /* end switch(padding) */
 
-		CRYPTO_THREAD_write_lock(cpriv->rwlock);
-		rv = CRYPTOKI_call(ctx,
-			C_SignInit(spriv->session, &mechanism, kpriv->object));
-		if (!rv && kpriv->always_authenticate == CK_TRUE)
-			rv = pkcs11_authenticate(key);
-	}
+	if (pkcs11_get_session(slot, 0, &session))
+		return -1;
+
+	rv = CRYPTOKI_call(ctx,
+		C_SignInit(session, &mechanism, key->object));
+	if (!rv && key->always_authenticate == CK_TRUE)
+		rv = pkcs11_authenticate(key, session);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
-			C_Sign(spriv->session, (CK_BYTE_PTR)tbs, tbslen, sig, &size));
-	cpriv->sign_initialized = !rv && sig == NULL;
-	if (!cpriv->sign_initialized)
-		CRYPTO_THREAD_unlock(cpriv->rwlock);
+			C_Sign(session, (CK_BYTE_PTR)tbs, tbslen, sig, &size));
+	pkcs11_put_session(slot, session);
 #ifdef DEBUG
 	fprintf(stderr, "%s:%d C_SignInit or C_Sign rv=%d\n",
 		__FILE__, __LINE__, rv);
@@ -392,85 +419,84 @@ static int pkcs11_try_pkey_rsa_decrypt(EVP_PKEY_CTX *evp_pkey_ctx,
 {
 	EVP_PKEY *pkey;
 	RSA *rsa;
-	PKCS11_KEY *key;
-	int rv = 0;
+	int rv = 0, padding;
 	CK_ULONG size = *outlen;
-	PKCS11_SLOT *slot;
-	PKCS11_CTX *ctx;
-	PKCS11_KEY_private *kpriv;
-	PKCS11_SLOT_private *spriv;
-	PKCS11_CTX_private *cpriv;
+	PKCS11_OBJECT_private *key;
+	PKCS11_SLOT_private *slot;
+	PKCS11_CTX_private *ctx;
+	CK_SESSION_HANDLE session;
+	CK_MECHANISM mechanism;
+	CK_RSA_PKCS_OAEP_PARAMS oaep_params;
 
 #ifdef DEBUG
 	fprintf(stderr, "%s:%d pkcs11_try_pkey_rsa_decrypt() "
 		"out=%p *outlen=%lu in=%p inlen=%lu\n",
 		__FILE__, __LINE__, out, *outlen, in, inlen);
 #endif
+	/* RSA method has EVP_PKEY_FLAG_AUTOARGLEN set. OpenSSL core will handle
+	 * the size inquiry internally. */
+	 if (!out)
+		return -1;
+
 	pkey = EVP_PKEY_CTX_get0_pkey(evp_pkey_ctx);
-	if (pkey == NULL)
+	if (!pkey)
 		return -1;
-	rsa = EVP_PKEY_get0_RSA(pkey);
-	if (rsa == NULL)
+	rsa = (RSA *)EVP_PKEY_get0_RSA(pkey);
+	if (!rsa)
 		return -1;
+
 	key = pkcs11_get_ex_data_rsa(rsa);
-	if (check_key_fork(key) < 0)
-		return -1;
-	slot = KEY2SLOT(key);
-	ctx = KEY2CTX(key);
-	kpriv = PRIVKEY(key);
-	spriv = PRIVSLOT(slot);
-	cpriv = PRIVCTX(ctx);
-
-	if (evp_pkey_ctx == NULL)
+	if (check_object_fork(key) < 0)
 		return -1;
 
-	if (!cpriv->decrypt_initialized) {
-		int padding;
-		CK_MECHANISM mechanism;
-		CK_RSA_PKCS_OAEP_PARAMS oaep_params;
+	slot = key->slot;
+	ctx = slot->ctx;
 
-		memset(&mechanism, 0, sizeof mechanism);
-		EVP_PKEY_CTX_get_rsa_padding(evp_pkey_ctx, &padding);
-		switch (padding) {
-		case RSA_PKCS1_OAEP_PADDING:
+	if (!evp_pkey_ctx)
+		return -1;
+
+	memset(&mechanism, 0, sizeof mechanism);
+	EVP_PKEY_CTX_get_rsa_padding(evp_pkey_ctx, &padding);
+	switch (padding) {
+	case RSA_PKCS1_OAEP_PADDING:
 #ifdef DEBUG
-			fprintf(stderr, "%s:%d padding=RSA_PKCS1_OAEP_PADDING\n",
-				__FILE__, __LINE__);
-#endif
-			if (pkcs11_params_oaep(&oaep_params, evp_pkey_ctx) < 0)
-				return -1;
-			mechanism.mechanism = CKM_RSA_PKCS_OAEP;
-			mechanism.pParameter = &oaep_params;
-			mechanism.ulParameterLen = sizeof oaep_params;
-			break;
-		case CKM_RSA_PKCS:
-#ifdef DEBUG
-			fprintf(stderr, "%s:%d padding=CKM_RSA_PKCS\n",
-				__FILE__, __LINE__);
-#endif
-			mechanism.pParameter = NULL;
-			mechanism.ulParameterLen = 0;
-			break;
-		default:
-#ifdef DEBUG
-			fprintf(stderr, "%s:%d unsupported padding: %d\n",
-				__FILE__, __LINE__, padding);
+		fprintf(stderr, "%s:%d padding=RSA_PKCS1_OAEP_PADDING\n",
+			__FILE__, __LINE__);
 #endif
+		if (pkcs11_params_oaep(&oaep_params, evp_pkey_ctx) < 0)
 			return -1;
-		} /* end switch(padding) */
+		mechanism.mechanism = CKM_RSA_PKCS_OAEP;
+		mechanism.pParameter = &oaep_params;
+		mechanism.ulParameterLen = sizeof oaep_params;
+		break;
+	case RSA_PKCS1_PADDING:
+#ifdef DEBUG
+		fprintf(stderr, "%s:%d padding=RSA_PKCS1_PADDING\n",
+			__FILE__, __LINE__);
+#endif
+		mechanism.mechanism = CKM_RSA_PKCS;
+		mechanism.pParameter = NULL;
+		mechanism.ulParameterLen = 0;
+		break;
+	default:
+#ifdef DEBUG
+		fprintf(stderr, "%s:%d unsupported padding: %d\n",
+			__FILE__, __LINE__, padding);
+#endif
+		return -1;
+	} /* end switch(padding) */
 
-		CRYPTO_THREAD_write_lock(cpriv->rwlock);
-		rv = CRYPTOKI_call(ctx,
-			C_DecryptInit(spriv->session, &mechanism, kpriv->object));
-		if (!rv && kpriv->always_authenticate == CK_TRUE)
-			rv = pkcs11_authenticate(key);
-	}
+	if (pkcs11_get_session(slot, 0, &session))
+		return -1;
+
+	rv = CRYPTOKI_call(ctx,
+		C_DecryptInit(session, &mechanism, key->object));
+	if (!rv && key->always_authenticate == CK_TRUE)
+		rv = pkcs11_authenticate(key, session);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
-			C_Decrypt(spriv->session, (CK_BYTE_PTR)in, inlen, out, &size));
-	cpriv->decrypt_initialized = !rv && out == NULL;
-	if (!cpriv->decrypt_initialized)
-		CRYPTO_THREAD_unlock(cpriv->rwlock);
+			C_Decrypt(session, (CK_BYTE_PTR)in, inlen, out, &size));
+	pkcs11_put_session(slot, session);
 #ifdef DEBUG
 	fprintf(stderr, "%s:%d C_DecryptInit or C_Decrypt rv=%d\n",
 		__FILE__, __LINE__, rv);
@@ -508,6 +534,11 @@ static EVP_PKEY_METHOD *pkcs11_pkey_method_rsa()
 	new_meth = EVP_PKEY_meth_new(EVP_PKEY_RSA,
 		EVP_PKEY_FLAG_AUTOARGLEN);
 
+#ifdef EVP_PKEY_FLAG_DYNAMIC
+	/* do not allow OpenSSL to free this object */
+	new_meth->flags &= ~EVP_PKEY_FLAG_DYNAMIC;
+#endif
+
 	EVP_PKEY_meth_copy(new_meth, orig_meth);
 
 	EVP_PKEY_meth_set_sign(new_meth,
@@ -526,16 +557,15 @@ static int pkcs11_try_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 {
 	EVP_PKEY *pkey;
 	EC_KEY *eckey;
-	PKCS11_KEY *key;
-	int rv = 0;
+	int rv = CKR_GENERAL_ERROR;
 	CK_ULONG size = *siglen;
-	PKCS11_SLOT *slot;
-	PKCS11_CTX *ctx;
-	PKCS11_KEY_private *kpriv;
-	PKCS11_SLOT_private *spriv;
-	PKCS11_CTX_private *cpriv;
+	PKCS11_OBJECT_private *key;
+	PKCS11_SLOT_private *slot;
+	PKCS11_CTX_private *ctx;
+	CK_SESSION_HANDLE session;
 	const EVP_MD *sig_md;
 	ECDSA_SIG *ossl_sig;
+	CK_MECHANISM mechanism;
 
 #ifdef DEBUG
 	fprintf(stderr, "%s:%d pkcs11_try_pkey_ec_sign() "
@@ -544,59 +574,57 @@ static int pkcs11_try_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 #endif
 
 	ossl_sig = ECDSA_SIG_new();
-	if (ossl_sig == NULL)
-		return-1;
+	if (!ossl_sig)
+		goto error;
 
 	pkey = EVP_PKEY_CTX_get0_pkey(evp_pkey_ctx);
-	if (pkey == NULL)
-		return -1;
+	if (!pkey)
+		goto error;
 
 	eckey = (EC_KEY *)EVP_PKEY_get0_EC_KEY(pkey);
-	if (eckey == NULL)
-		return -1;
+	if (!eckey)
+		goto error;
+
+	if (!sig) {
+		*siglen = (size_t)ECDSA_size(eckey);
+		rv = CKR_OK;
+		goto error;
+	}
 
 	if (*siglen < (size_t)ECDSA_size(eckey))
-		return -1;
+		goto error;
 
 	key = pkcs11_get_ex_data_ec(eckey);
-	if (check_key_fork(key) < 0)
-		return -1;
+	if (check_object_fork(key) < 0)
+		goto error;
 
-	slot = KEY2SLOT(key);
-	ctx = KEY2CTX(key);
-	kpriv = PRIVKEY(key);
-	spriv = PRIVSLOT(slot);
-	cpriv = PRIVCTX(ctx);
+	slot = key->slot;
+	ctx = slot->ctx;
 
-	if (evp_pkey_ctx == NULL)
-		return -1;
+	if (!evp_pkey_ctx)
+		goto error;
 
 	if (EVP_PKEY_CTX_get_signature_md(evp_pkey_ctx, &sig_md) <= 0)
-		return -1;
+		goto error;
 
 	if (tbslen < (size_t)EVP_MD_size(sig_md))
+		goto error;
+
+	rv = 0;
+	memset(&mechanism, 0, sizeof mechanism);
+	mechanism.mechanism = CKM_ECDSA;
+
+	if (pkcs11_get_session(slot, 0, &session))
 		return -1;
-
-	if (!cpriv->sign_initialized) {
-		int padding;
-		CK_MECHANISM mechanism;
-		memset(&mechanism, 0, sizeof mechanism);
-
-		mechanism.mechanism = CKM_ECDSA;
-
-		CRYPTO_THREAD_write_lock(cpriv->rwlock);
-		rv = CRYPTOKI_call(ctx,
-			C_SignInit(spriv->session, &mechanism, kpriv->object));
-		if (!rv && kpriv->always_authenticate == CK_TRUE)
-			rv = pkcs11_authenticate(key);
-	}
+	rv = CRYPTOKI_call(ctx,
+		C_SignInit(session, &mechanism, key->object));
+	if (!rv && key->always_authenticate == CK_TRUE)
+		rv = pkcs11_authenticate(key, session);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
-			C_Sign(spriv->session, (CK_BYTE_PTR)tbs, tbslen, sig, &size));
+			C_Sign(session, (CK_BYTE_PTR)tbs, tbslen, sig, &size));
+	pkcs11_put_session(slot, session);
 
-	cpriv->sign_initialized = !rv && sig == NULL;
-	if (!cpriv->sign_initialized)
-		CRYPTO_THREAD_unlock(cpriv->rwlock);
 #ifdef DEBUG
 	fprintf(stderr, "%s:%d C_SignInit or C_Sign rv=%d\n",
 		__FILE__, __LINE__, rv);
@@ -606,7 +634,7 @@ static int pkcs11_try_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 		BIGNUM *r = BN_bin2bn(sig, size/2, NULL);
 		BIGNUM *s = BN_bin2bn(sig + size/2, size/2, NULL);
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER >= 0x3050000fL )
 		ECDSA_SIG_set0(ossl_sig, r, s);
 #else
 		BN_free(ossl_sig->r);
@@ -617,6 +645,7 @@ static int pkcs11_try_pkey_ec_sign(EVP_PKEY_CTX *evp_pkey_ctx,
 		*siglen = i2d_ECDSA_SIG(ossl_sig, &sig);
 	}
 
+error:
 	ECDSA_SIG_free(ossl_sig);
 
 	if (rv != CKR_OK)
@@ -645,8 +674,12 @@ static EVP_PKEY_METHOD *pkcs11_pkey_method_ec()
 	EVP_PKEY_meth_get_sign(orig_meth,
 		&orig_pkey_ec_sign_init, &orig_pkey_ec_sign);
 
-	new_meth = EVP_PKEY_meth_new(EVP_PKEY_EC,
-		EVP_PKEY_FLAG_AUTOARGLEN);
+	new_meth = EVP_PKEY_meth_new(EVP_PKEY_EC, 0);
+
+#ifdef EVP_PKEY_FLAG_DYNAMIC
+	/* do not allow OpenSSL to free this object */
+	new_meth->flags &= ~EVP_PKEY_FLAG_DYNAMIC;
+#endif
 
 	EVP_PKEY_meth_copy(new_meth, orig_meth);
 
@@ -680,17 +713,17 @@ int PKCS11_pkey_meths(ENGINE *e, EVP_PKEY_METHOD **pmeth,
 	/* get the EVP_PKEY_METHOD */
 	switch (nid) {
 	case EVP_PKEY_RSA:
-		if (pkey_method_rsa == NULL)
+		if (!pkey_method_rsa)
 			pkey_method_rsa = pkcs11_pkey_method_rsa();
-		if (pkey_method_rsa == NULL)
+		if (!pkey_method_rsa)
 			return 0;
 		*pmeth = pkey_method_rsa;
 		return 1; /* success */
 #ifndef OPENSSL_NO_EC
 	case EVP_PKEY_EC:
-		if (pkey_method_ec == NULL)
+		if (!pkey_method_ec)
 			pkey_method_ec = pkcs11_pkey_method_ec();
-		if (pkey_method_ec == NULL)
+		if (!pkey_method_ec)
 			return 0;
 		*pmeth = pkey_method_ec;
 		return 1; /* success */
