@@ -29,13 +29,16 @@
 
 static int rsa_ex_index = 0;
 
-static RSA *pkcs11_rsa(PKCS11_KEY *key)
+static RSA *pkcs11_rsa(PKCS11_OBJECT_private *key)
 {
-	EVP_PKEY *evp_key = pkcs11_get_key(key, key->isPrivate);
+	EVP_PKEY *evp_key = pkcs11_get_key(key, key->object_class);
 	RSA *rsa;
 	if (!evp_key)
 		return NULL;
-	rsa = EVP_PKEY_get0_RSA(evp_key);
+	rsa = (RSA *)EVP_PKEY_get0_RSA(evp_key);
+	/* Danger: this assumes evp_key returned above has at least reference
+	 * count of 2. Which is true in current code as long as key->object_class
+	 * is used for the object_class. */
 	EVP_PKEY_free(evp_key);
 	return rsa;
 }
@@ -43,7 +46,7 @@ static RSA *pkcs11_rsa(PKCS11_KEY *key)
 /* PKCS#1 v1.5 RSA signature */
 /* TODO: remove this function in libp11 0.5.0 */
 int pkcs11_sign(int type, const unsigned char *m, unsigned int m_len,
-		unsigned char *sigret, unsigned int *siglen, PKCS11_KEY *key)
+		unsigned char *sigret, unsigned int *siglen, PKCS11_OBJECT_private *key)
 {
 	RSA *rsa = pkcs11_rsa(key);
 	if (!rsa)
@@ -76,14 +79,13 @@ static int pkcs11_mechanism(CK_MECHANISM *mechanism, const int padding)
 /* OpenSSL assumes that the output buffer is always big enough */
 int pkcs11_private_encrypt(int flen,
 		const unsigned char *from, unsigned char *to,
-		PKCS11_KEY *key, int padding)
+		PKCS11_OBJECT_private *key, int padding)
 {
-	PKCS11_SLOT *slot = KEY2SLOT(key);
-	PKCS11_CTX *ctx = KEY2CTX(key);
-	PKCS11_KEY_private *kpriv = PRIVKEY(key);
-	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
+	PKCS11_SLOT_private *slot = key->slot;
+	PKCS11_CTX_private *ctx = slot->ctx;
 	CK_MECHANISM mechanism;
 	CK_ULONG size;
+	CK_SESSION_HANDLE session;
 	int rv;
 
 	size = pkcs11_get_key_size(key);
@@ -91,26 +93,28 @@ int pkcs11_private_encrypt(int flen,
 	if (pkcs11_mechanism(&mechanism, padding) < 0)
 		return -1;
 
-	CRYPTO_THREAD_write_lock(PRIVCTX(ctx)->rwlock);
+	if (pkcs11_get_session(slot, 0, &session))
+		return -1;
+
 	/* Try signing first, as applications are more likely to use it */
 	rv = CRYPTOKI_call(ctx,
-		C_SignInit(spriv->session, &mechanism, kpriv->object));
-	if (!rv && kpriv->always_authenticate == CK_TRUE)
-		rv = pkcs11_authenticate(key);
+		C_SignInit(session, &mechanism, key->object));
+	if (!rv && key->always_authenticate == CK_TRUE)
+		rv = pkcs11_authenticate(key, session);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
-			C_Sign(spriv->session, (CK_BYTE *)from, flen, to, &size));
+			C_Sign(session, (CK_BYTE *)from, flen, to, &size));
 	if (rv == CKR_KEY_FUNCTION_NOT_PERMITTED) {
 		/* OpenSSL may use it for encryption rather than signing */
 		rv = CRYPTOKI_call(ctx,
-			C_EncryptInit(spriv->session, &mechanism, kpriv->object));
-		if (!rv && kpriv->always_authenticate == CK_TRUE)
-			rv = pkcs11_authenticate(key);
+			C_EncryptInit(session, &mechanism, key->object));
+		if (!rv && key->always_authenticate == CK_TRUE)
+			rv = pkcs11_authenticate(key, session);
 		if (!rv)
 			rv = CRYPTOKI_call(ctx,
-				C_Encrypt(spriv->session, (CK_BYTE *)from, flen, to, &size));
+				C_Encrypt(session, (CK_BYTE *)from, flen, to, &size));
 	}
-	CRYPTO_THREAD_unlock(PRIVCTX(ctx)->rwlock);
+	pkcs11_put_session(slot, session);
 
 	if (rv) {
 		CKRerr(CKR_F_PKCS11_PRIVATE_ENCRYPT, rv);
@@ -122,12 +126,11 @@ int pkcs11_private_encrypt(int flen,
 
 /* RSA private key decryption */
 int pkcs11_private_decrypt(int flen, const unsigned char *from, unsigned char *to,
-		PKCS11_KEY *key, int padding)
+		PKCS11_OBJECT_private *key, int padding)
 {
-	PKCS11_SLOT *slot = KEY2SLOT(key);
-	PKCS11_CTX *ctx = KEY2CTX(key);
-	PKCS11_KEY_private *kpriv = PRIVKEY(key);
-	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
+	PKCS11_SLOT_private *slot = key->slot;
+	PKCS11_CTX_private *ctx = slot->ctx;
+	CK_SESSION_HANDLE session;
 	CK_MECHANISM mechanism;
 	CK_ULONG size = flen;
 	CK_RV rv;
@@ -135,16 +138,18 @@ int pkcs11_private_decrypt(int flen, const unsigned char *from, unsigned char *t
 	if (pkcs11_mechanism(&mechanism, padding) < 0)
 		return -1;
 
-	CRYPTO_THREAD_write_lock(PRIVCTX(ctx)->rwlock);
+	if (pkcs11_get_session(slot, 0, &session))
+		return -1;
+
 	rv = CRYPTOKI_call(ctx,
-		C_DecryptInit(spriv->session, &mechanism, kpriv->object));
-	if (!rv && kpriv->always_authenticate == CK_TRUE)
-		rv = pkcs11_authenticate(key);
+		C_DecryptInit(session, &mechanism, key->object));
+	if (!rv && key->always_authenticate == CK_TRUE)
+		rv = pkcs11_authenticate(key, session);
 	if (!rv)
 		rv = CRYPTOKI_call(ctx,
-			C_Decrypt(spriv->session, (CK_BYTE *)from, size,
+			C_Decrypt(session, (CK_BYTE *)from, size,
 				(CK_BYTE_PTR)to, &size));
-	CRYPTO_THREAD_unlock(PRIVCTX(ctx)->rwlock);
+	pkcs11_put_session(slot, session);
 
 	if (rv) {
 		CKRerr(CKR_F_PKCS11_PRIVATE_DECRYPT, rv);
@@ -156,7 +161,7 @@ int pkcs11_private_decrypt(int flen, const unsigned char *from, unsigned char *t
 
 /* TODO: remove this function in libp11 0.5.0 */
 int pkcs11_verify(int type, const unsigned char *m, unsigned int m_len,
-		unsigned char *signature, unsigned int siglen, PKCS11_KEY *key)
+		unsigned char *signature, unsigned int siglen, PKCS11_OBJECT_private *key)
 {
 	(void)type;
 	(void)m;
@@ -173,19 +178,27 @@ int pkcs11_verify(int type, const unsigned char *m, unsigned int m_len,
 /*
  * Get RSA key material
  */
-static RSA *pkcs11_get_rsa(PKCS11_KEY *key)
+static RSA *pkcs11_get_rsa(PKCS11_OBJECT_private *key)
 {
+	CK_OBJECT_CLASS class_public_key = CKO_PUBLIC_KEY;
+	PKCS11_SLOT_private *slot = key->slot;
+	PKCS11_CTX_private *ctx = slot->ctx;
+	PKCS11_OBJECT_private *pubkey;
+	PKCS11_TEMPLATE tmpl = {0};
+	CK_OBJECT_HANDLE object = key->object;
+	CK_SESSION_HANDLE session;
 	RSA *rsa;
-	PKCS11_KEY *keys;
-	unsigned int i, count;
 	BIGNUM *rsa_n = NULL, *rsa_e = NULL;
 
-	/* Retrieve the modulus */
-	if (key_getattr_bn(key, CKA_MODULUS, &rsa_n))
+	if (pkcs11_get_session(slot, 0, &session))
 		return NULL;
 
+	/* Retrieve the modulus */
+	if (pkcs11_getattr_bn(ctx, session, object, CKA_MODULUS, &rsa_n))
+		goto failure;
+
 	/* Retrieve the public exponent */
-	if (!key_getattr_bn(key, CKA_PUBLIC_EXPONENT, &rsa_e)) {
+	if (!pkcs11_getattr_bn(ctx, session, object, CKA_PUBLIC_EXPONENT, &rsa_e)) {
 		if (!BN_is_zero(rsa_e)) /* A valid public exponent */
 			goto success;
 		BN_clear_free(rsa_e);
@@ -194,18 +207,15 @@ static RSA *pkcs11_get_rsa(PKCS11_KEY *key)
 
 	/* The public exponent was not found in the private key:
 	 * retrieve it from the corresponding public key */
-	if (!PKCS11_enumerate_public_keys(KEY2TOKEN(key), &keys, &count)) {
-		for (i = 0; i < count; i++) {
-			BIGNUM *pubmod = NULL;
-			if (!key_getattr_bn(&keys[i], CKA_MODULUS, &pubmod)) {
-				int found = BN_cmp(rsa_n, pubmod) == 0;
-				BN_clear_free(pubmod);
-				if (found && !key_getattr_bn(&keys[i],
-						CKA_PUBLIC_EXPONENT, &rsa_e))
-					goto success;
-			}
-		}
+	pkcs11_addattr_var(&tmpl, CKA_CLASS, class_public_key);
+	pkcs11_addattr_bn(&tmpl, CKA_MODULUS, rsa_n);
+	pubkey = pkcs11_object_from_template(slot, session, &tmpl);
+	if (pubkey && !pkcs11_getattr_bn(ctx, session, pubkey->object,
+			CKA_PUBLIC_EXPONENT, &rsa_e)) {
+		pkcs11_object_free(pubkey);
+		goto success;
 	}
+	pkcs11_object_free(pubkey);
 
 	/* Last resort: use the most common default */
 	rsa_e = BN_new();
@@ -213,6 +223,7 @@ static RSA *pkcs11_get_rsa(PKCS11_KEY *key)
 		goto success;
 
 failure:
+	pkcs11_put_session(slot, session);
 	if (rsa_n)
 		BN_clear_free(rsa_n);
 	if (rsa_e)
@@ -220,10 +231,11 @@ failure:
 	return NULL;
 
 success:
+	pkcs11_put_session(slot, session);
 	rsa = RSA_new();
 	if (!rsa)
 		goto failure;
-#if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER >= 0x3050000fL )
 	RSA_set0_key(rsa, rsa_n, rsa_e, NULL);
 #else
 	rsa->n = rsa_n;
@@ -232,33 +244,20 @@ success:
 	return rsa;
 }
 
-PKCS11_KEY *pkcs11_get_ex_data_rsa(const RSA *rsa)
+PKCS11_OBJECT_private *pkcs11_get_ex_data_rsa(const RSA *rsa)
 {
 	return RSA_get_ex_data(rsa, rsa_ex_index);
 }
 
-static void pkcs11_set_ex_data_rsa(RSA *rsa, PKCS11_KEY *key)
+static void pkcs11_set_ex_data_rsa(RSA *rsa, PKCS11_OBJECT_private *key)
 {
 	RSA_set_ex_data(rsa, rsa_ex_index, key);
 }
 
-static void pkcs11_update_ex_data_rsa(PKCS11_KEY *key)
-{
-	EVP_PKEY *evp = key->evp_key;
-	RSA *rsa;
-	if (!evp)
-		return;
-	if (EVP_PKEY_base_id(evp) != EVP_PKEY_RSA)
-		return;
-
-	rsa = EVP_PKEY_get1_RSA(evp);
-	pkcs11_set_ex_data_rsa(rsa, key);
-	RSA_free(rsa);
-}
 /*
  * Build an EVP_PKEY object
  */
-static EVP_PKEY *pkcs11_get_evp_key_rsa(PKCS11_KEY *key)
+static EVP_PKEY *pkcs11_get_evp_key_rsa(PKCS11_OBJECT_private *key)
 {
 	EVP_PKEY *pk;
 	RSA *rsa;
@@ -271,11 +270,9 @@ static EVP_PKEY *pkcs11_get_evp_key_rsa(PKCS11_KEY *key)
 		RSA_free(rsa);
 		return NULL;
 	}
-	EVP_PKEY_set1_RSA(pk, rsa); /* Also increments the rsa ref count */
-
-	if (key->isPrivate) {
+	if (key->object_class == CKO_PRIVATE_KEY) {
 		RSA_set_method(rsa, PKCS11_get_rsa_method());
-#if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER >= 0x3050000fL )
 		RSA_set_flags(rsa, RSA_FLAG_EXT_PKEY);
 #else
 		rsa->flags |= RSA_FLAG_EXT_PKEY;
@@ -289,19 +286,21 @@ static EVP_PKEY *pkcs11_get_evp_key_rsa(PKCS11_KEY *key)
 	rsa->flags |= RSA_FLAG_SIGN_VER;
 #endif
 	pkcs11_set_ex_data_rsa(rsa, key);
+
+	EVP_PKEY_set1_RSA(pk, rsa); /* Also increments the rsa ref count */
 	RSA_free(rsa); /* Drops our reference to it */
 	return pk;
 }
 
 /* TODO: remove this function in libp11 0.5.0 */
-int pkcs11_get_key_modulus(PKCS11_KEY *key, BIGNUM **bn)
+int pkcs11_get_key_modulus(PKCS11_OBJECT_private *key, BIGNUM **bn)
 {
 	RSA *rsa = pkcs11_rsa(key);
 	const BIGNUM *rsa_n;
 
 	if (!rsa)
 		return 0;
-#if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER >= 0x3050000fL )
 	RSA_get0_key(rsa, &rsa_n, NULL, NULL);
 #else
 	rsa_n=rsa->n;
@@ -311,14 +310,14 @@ int pkcs11_get_key_modulus(PKCS11_KEY *key, BIGNUM **bn)
 }
 
 /* TODO: remove this function in libp11 0.5.0 */
-int pkcs11_get_key_exponent(PKCS11_KEY *key, BIGNUM **bn)
+int pkcs11_get_key_exponent(PKCS11_OBJECT_private *key, BIGNUM **bn)
 {
 	RSA *rsa = pkcs11_rsa(key);
 	const BIGNUM *rsa_e;
 
 	if (!rsa)
 		return 0;
-#if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER >= 0x3050000fL )
 	RSA_get0_key(rsa, NULL, &rsa_e, NULL);
 #else
 	rsa_e=rsa->e;
@@ -328,7 +327,7 @@ int pkcs11_get_key_exponent(PKCS11_KEY *key, BIGNUM **bn)
 }
 
 /* TODO: make this function static in libp11 0.5.0 */
-int pkcs11_get_key_size(PKCS11_KEY *key)
+int pkcs11_get_key_size(PKCS11_OBJECT_private *key)
 {
 	RSA *rsa = pkcs11_rsa(key);
 	if (!rsa)
@@ -336,7 +335,7 @@ int pkcs11_get_key_size(PKCS11_KEY *key)
 	return RSA_size(rsa);
 }
 
-#if OPENSSL_VERSION_NUMBER < 0x10100005L || defined(LIBRESSL_VERSION_NUMBER)
+#if ( ( defined (OPENSSL_VERSION_NUMBER) && OPENSSL_VERSION_NUMBER < 0x10100005L ) || ( defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x3020199L ) )
 
 int (*RSA_meth_get_priv_enc(const RSA_METHOD *meth))
 		(int flen, const unsigned char *from,
@@ -362,32 +361,36 @@ static int (*RSA_meth_get_finish(const RSA_METHOD *meth)) (RSA *rsa)
 static int pkcs11_rsa_priv_dec_method(int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding)
 {
-	PKCS11_KEY *key = pkcs11_get_ex_data_rsa(rsa);
+	PKCS11_OBJECT_private *key = pkcs11_get_ex_data_rsa(rsa);
 	int (*priv_dec) (int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding);
-	if (check_key_fork(key) < 0) {
+	if (check_object_fork(key) < 0) {
 		priv_dec = RSA_meth_get_priv_dec(RSA_get_default_method());
 		return priv_dec(flen, from, to, rsa, padding);
 	}
-	return PKCS11_private_decrypt(flen, from, to, key, padding);
+	return pkcs11_private_decrypt(flen, from, to, key, padding);
 }
 
 static int pkcs11_rsa_priv_enc_method(int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding)
 {
-	PKCS11_KEY *key = pkcs11_get_ex_data_rsa(rsa);
+	PKCS11_OBJECT_private *key = pkcs11_get_ex_data_rsa(rsa);
 	int (*priv_enc) (int flen, const unsigned char *from,
 		unsigned char *to, RSA *rsa, int padding);
-	if (check_key_fork(key) < 0) {
+	if (check_object_fork(key) < 0) {
 		priv_enc = RSA_meth_get_priv_enc(RSA_get_default_method());
 		return priv_enc(flen, from, to, rsa, padding);
 	}
-	return PKCS11_private_encrypt(flen, from, to, key, padding);
+	return pkcs11_private_encrypt(flen, from, to, key, padding);
 }
 
 static int pkcs11_rsa_free_method(RSA *rsa)
 {
-	RSA_set_ex_data(rsa, rsa_ex_index, NULL);
+	PKCS11_OBJECT_private *key = pkcs11_get_ex_data_rsa(rsa);
+	if (key) {
+		pkcs11_set_ex_data_rsa(rsa, NULL);
+		pkcs11_object_free(key);
+	}
 	int (*orig_rsa_free_method)(RSA *rsa) =
 		RSA_meth_get_finish(RSA_get_default_method());
 	if (orig_rsa_free_method) {
@@ -508,10 +511,9 @@ void PKCS11_rsa_method_free(void)
 	free_rsa_ex_index();
 }
 
-PKCS11_KEY_ops pkcs11_rsa_ops = {
+PKCS11_OBJECT_ops pkcs11_rsa_ops = {
 	EVP_PKEY_RSA,
 	pkcs11_get_evp_key_rsa,
-	pkcs11_update_ex_data_rsa
 };
 
 /* vim: set noexpandtab: */
